#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include "IOTCDevice.h"
#include "RDTServer.h"
#include "common.h"
#include <windows.h>

#define RDT_USERNAME "admin"
#define RDT_PASSWORD "testpwd"

int gFd;
static char gFn[64];

struct channelSendData_s
{
	int SID;
	unsigned char CHID;
};

static void usleep(long long usec) {
    HANDLE timer;
    LARGE_INTEGER interval;
    interval.QuadPart = -(10 * usec);

    timer = CreateWaitableTimer(NULL, TRUE, NULL);
    SetWaitableTimer(timer, &interval, 0, NULL, NULL, 0);
    WaitForSingleObject(timer, INFINITE);
    CloseHandle(timer);
}

void WINAPI RDT_Server_Authentication(const char *cszAccount, char *cszPassword, uint32_t nPasswordMaxLength, const void *pArg) {
    printf("[%s] cszAccount=%s, nPasswordMaxLength=%d, pArg=%s\n", __func__, cszAccount, nPasswordMaxLength, (char *)pArg);
    if (strcmp(cszAccount, RDT_USERNAME) == 0) {
        strcpy(cszPassword, RDT_PASSWORD);
    } else {
        printf("unknow account, reject connect\n");
        // do nothing to reject connect
    }
}

void WINAPI thread_channelSendData(void * arg)
{
	struct channelSendData_s * chSendArg = (struct channelSendData_s *) arg;
	int SID = chSendArg->SID;
	unsigned char CHID = chSendArg->CHID;
	struct st_SInfoEx sInfo;
	char buf[MAX_BUF_SIZE];

	FILE *fp = fopen(gFn, "rb");
	if(fp == NULL)
	{
		printf("FOpen file[%s] errno[%d]!\n", gFn, errno);
		return;
	}

    sInfo.size = sizeof(struct st_SInfoEx);
	IOTC_Session_Check_Ex(SID, &sInfo);

    char *args = strdup("test_args");
    int RDT_ID = RDT_Server_Create(SID, CHID, RDT_Server_Authentication, args, RDT_WAIT_TIMEMS);
    if (RDT_ID < 0) {
        printf("RDT_Create failed[%d]!!\n", RDT_ID);
        IOTC_Session_Close(SID);
        fclose(fp);
        return;
	}

	printf("RDT_ID = %d\n", RDT_ID);
#if 0
	int ret = RDT_Read(RDT_ID, buf, 1024, RDT_WAIT_TIMEMS);
	if(ret < 0)
	{
		printf("RDT rcv cmd error[%d]!\n", ret);
		fclose(fp);
		return;
	}
	int requestCnt = atoi(buf);
#endif
	int ret = 0;
	struct stat fStat;
	fstat(gFd, &fStat);
	struct st_RDT_Status rdtStatus;
	printf("[%d]file size[%lu]\n", RDT_ID, fStat.st_size);
	sprintf(buf, "%lu", fStat.st_size);
	ret =  RDT_Write(RDT_ID, buf, strlen(buf));
	if(ret < 0)
	{
		printf("[%d]RDT send file size error[%d]!\n", RDT_ID, ret);
		fclose(fp);
		return;
	}
	ret = RDT_Read(RDT_ID, buf, 1024, RDT_WAIT_TIMEMS);
	if(ret < 0 && strcmp(buf, START_STRING) != 0)
	{
		printf("[%d]RDT rcv start to send file error[%d]!\n", RDT_ID, ret);
		fclose(fp);
		return;
	}

	while(1)
	{
		ret = fread(buf, 1, MAX_BUF_SIZE, fp);
		if(ret <= 0) break;
		ret = RDT_Write(RDT_ID, buf, ret);
		if(ret < 0)
		{
			printf("[%d]RDT send file data error[%d]!\n", RDT_ID, ret);
			return;
		}

		if((ret = RDT_Status_Check(RDT_ID, &rdtStatus)) == RDT_ER_NoERROR)
		{
			if(sInfo.Mode == 2)
			{
				if(rdtStatus.BufSizeInSendQueue > 1024000)
					usleep(50000);
			}
			else
			{
				if(rdtStatus.BufSizeInSendQueue > 512000)
					usleep(100000);
				//Sleep(1000);
			}
		}
		else
		{
			printf("[%d]RDT status check error[%d]!\n", RDT_ID, ret);
			return;
		}
	}

	fseek(fp, 0, SEEK_SET);

	printf("[%d]***RDT_Destroy enter\n", RDT_ID);
	RDT_Destroy(RDT_ID);
	fclose(fp);
	IOTC_Session_Channel_OFF(SID, CHID);
	
	printf("thread_Send exit[%d]\n", RDT_ID);

	return;

}

void WINAPI thread_Send(void *arg)
{
	struct channelSendData_s argThread1;
	struct channelSendData_s argThread2;
	struct channelSendData_s argThread3;
	HANDLE channelSendDataThread1;
	HANDLE channelSendDataThread2;
	HANDLE channelSendDataThread3;
	int SID = *(int *)arg;
	free(arg);


	memset(&argThread1, 0, sizeof(struct channelSendData_s));
	memset(&argThread2, 0, sizeof(struct channelSendData_s));
	memset(&argThread3, 0, sizeof(struct channelSendData_s));

	argThread1.SID = argThread2.SID = argThread3.SID = SID;
	argThread1.CHID = 0;
	argThread2.CHID = 1;
	argThread3.CHID = 2;

    channelSendDataThread1 = CreateThread(NULL, 0, thread_channelSendData, (void*)&argThread1, 0, NULL);
    channelSendDataThread2 = CreateThread(NULL, 0, thread_channelSendData, (void*)&argThread2, 0, NULL);
    channelSendDataThread3 = CreateThread(NULL, 0, thread_channelSendData, (void*)&argThread3, 0, NULL);

	WaitForSingleObject(channelSendDataThread1, INFINITE);
	WaitForSingleObject(channelSendDataThread2, INFINITE);
	WaitForSingleObject(channelSendDataThread3, INFINITE);
	CloseHandle(channelSendDataThread1);
	CloseHandle(channelSendDataThread2);
	CloseHandle(channelSendDataThread3);
	
	while(1)
	{
		struct st_SInfoEx info;
        info.size = sizeof(struct st_SInfoEx);
		if( IOTC_Session_Check_Ex(SID, &info) < 0)
		{
			IOTC_Session_Close(SID);
			break;
		}
		Sleep(1000);
	}

	return;
}

void WINAPI _thread_Login(void *arg)
{
	int ret;
	char *UID = (char *)arg;
	while(1)
	{
		ret = IOTC_Device_Login(UID, "TUTK", "1234");
		printf("     Calling IOTC_Device_Login() ret = %d\n", ret);
		if(ret == IOTC_ER_NoERROR)
			break;
		else
			Sleep(60000);
	}

	return;
}

#define MAX_CLIENT_NUM 8

int main(int argc, char *argv[])
{
	if(argc < 5)
	{
		printf("%s <LicenseKey> <UID> <Filename> <MaxClientNum>\n",argv[0]);
		return 0;
	}
    printf("RDT Ver[%s]\n", RDT_GetRDTApiVersionString());

    char *license_key = (char *)argv[1];
    int ret = TUTK_SDK_Set_License_Key(license_key);
    if (ret != TUTK_ER_NoERROR)
    {
        printf("TUTK_SDK_Set_License_Key()=%d ,exit...!!\n", ret);
        return 0;
    }

	int maxClientNum = atoi(argv[4]);
	strcpy(gFn, argv[3]);
	printf("%s\n", gFn);
	gFd = open(gFn, O_RDONLY);
	if(gFd < 0)
	{
		printf("Open file[%s] error code[%d], %d!!\n", gFn, gFd, errno);
		exit(0);
	}

	ret = IOTC_Initialize2(0);
	if(ret != IOTC_ER_NoERROR)
	{
		printf("IOTC_Initialize error!!\n");
		return 0;
	}

	int rdtCh = RDT_Initialize();
	if(rdtCh <= 0)
	{
		printf("RDT_Initialize error!!\n");
		return 0;
	}

	char *UID = (char *)argv[2];
	HANDLE threadID_Login;
    threadID_Login = CreateThread(NULL, 0, _thread_Login, (void*)UID, 0, NULL);
	CloseHandle(threadID_Login);

	int SID = -1, clientCnt = 0;
	struct st_SInfoEx Sinfo;
	printf("start IOTC_Listen...\n");
	//unsigned long t;
	HANDLE Thread_ID[MAX_CLIENT_NUM];
	do {
		SID = IOTC_Listen(300000);

		if(SID > -1)
		{
            int *pSID = malloc(sizeof(int));
            if(pSID == NULL)
            {
                printf("[System error] - malloc failed\n");
                exit(0);
            }
            Sinfo.size = sizeof(struct st_SInfoEx);
			IOTC_Session_Check_Ex(SID, &Sinfo);
			char *mode[3] = {"P2P", "RLY", "LAN"};
			printf("Client from %s:%d Mode=%s\n",Sinfo.RemoteIP, Sinfo.RemotePort, mode[(int)Sinfo.Mode]);
			*pSID = SID;
            Thread_ID[clientCnt] = CreateThread(NULL, 0, thread_Send, (void*)pSID, 0, NULL);
			//if(clientCnt == 0) t = IOTC_GetTickCount();
			clientCnt++;
		}
		if(clientCnt == maxClientNum)
			break;
	}while(1);

	int i;
	for(i=0;i<maxClientNum;i++) {
		WaitForSingleObject(Thread_ID[i], INFINITE);
		CloseHandle(Thread_ID[i]);
	}

	//printf("RDT Send Data Cost[%lu ms]....\n", IOTC_GetTickCount()-t);
	RDT_DeInitialize();
	printf("RDT_DeInitialize OK\n");
	IOTC_DeInitialize();

	close(gFd);
	printf("Server exit!\n");

	return 0;
}


